import types
import sys
import inspect
sys.path.insert(0, 'google_appengine')  # This part is need to access to google.appengine.ext.db
import dev_appserver                    # what is needed to know about db.*Property datatypes
dev_appserver.fix_sys_path()            #
from google.appengine.ext import db     #
from google.appengine.api.users import User

#-----------
import taintConfig
if not hasattr(taintConfig,'Policy'):
    taintConfig.Policy = taintConfig.SimpleTags
#-----------

def propagate_func(original):
    def inner (*args, **kwargs):
        t = taintConfig.Policy()
        for a in args:
            t.join(collect_tags(a))
        for v in kwargs.values():
            t.join(collect_tags(v))
        r  = original(*args, **kwargs)
        if t:
            r = taint_aware(r, t)
        return r
    return inner

def propagate_method(method):
    def inner(self, *args, **kwargs):
        r = method(self, *args, **kwargs)
        t = taintConfig.Policy()
        for a in args:
            t.join(collect_tags(a))
        for v in kwargs.values():
            t.join(collect_tags(v))
        t.join(self.taints)
        return taint_aware(r, t)
    return inner

def taint_buildin(klass, methods=None):
    if not methods:
        methods = attributes(klass)

    class tbuildin(klass):
        def __new__(cls, *args, **kwargs):
            self = super(tbuildin, cls).__new__(cls, *args, **kwargs) #justificar analizar pq no init
            self.taints = taintConfig.Policy()
            self.originalType = klass

            # if any of the arguments is tainted, taint the object aswell

            for a in args:      # this CHUNK of code appears at least 3 times, refactor later
                self.taints.join(collect_tags(a))
            for v in kwargs.values():
                self.taints.join(collect_tags(v))

            return self

    d = klass.__dict__
    for name, attr in [(m, d[m]) for m in methods]:
        if inspect.ismethod(attr) or inspect.ismethoddescriptor(attr) or inspect.isfunction(attr):
            setattr(tbuildin, name, propagate_method(attr))
    # str has no __radd__ method
    if '__add__' in methods and '__radd__' not in methods:
        setattr(tbuildin, '__radd__', lambda self, other:
                                    tbuildin.__add__(tbuildin(other), self))
    # unicode __rmod__ returns NotImplemented
    if klass == unicode:
        setattr(tbuildin, '__rmod__',
                lambda self, other:
                    #print 'XXXXX'
                    tbuildin.__mod__(tbuildin(other), self)
                    #u'Called'
               )

    return tbuildin

class opaque_taint(object):
    '''every method in my_instance will return a value tainted with my_instance.taints'''
    def __init__(self, other):
        self.other = other
        self.taints = taintConfig.Policy()

    def __getattr__(self, name):
        if hasattr(self.other, name):
            func = getattr(self.other, name)
            if callable(func):
                return lambda *args, **kwargs: self._wrap(func, args, kwargs)
            return taint_aware(func,self.taints)
        raise AttributeError(name)

    def _wrap(self, func, args, kwargs):
        result = propagate_func(func)(*args, **kwargs)
        for a in args:
            self.taints.join(collect_tags(a))
        for v in kwargs.values():
            self.taints.join(collect_tags(v))
        return taint_aware(result,self.taints)

dont_override = set(['__repr__', '__cmp__', '__getattribute__', '__new__',
                     '__init__','__nonzero__', '__reduce__', '__reduce_ex__',
                     '__str__', '__int__', '__float__', '__unicode__'])

def attributes(klass):
    a = set(klass.__dict__.keys())
    return a - dont_override

str_methods = attributes(str)
unicode_methods = attributes(unicode)
int_methods = attributes(int)
long_methods = attributes(long)
float_methods = attributes(float)

STR = taint_buildin(str, str_methods)
UNICODE = taint_buildin(unicode, unicode_methods)
INT = taint_buildin(int, int_methods)
LONG = taint_buildin(long, long_methods)
FLOAT = taint_buildin(float, float_methods)

ignorableTypes = [
    types.NoneType,
    types.BooleanType,
    types.FileType,
    sys.stderr.__class__, #This is 'real' file. Since GAE captures types.FileType and wrappers it with a google.appengine.tools.dev_appserver.FakeFile class.
    types.NotImplementedType,
    db.datetime.datetime,
    User,
    ]

from cStringIO import InputType,OutputType

tclasses = {
    str: STR,
    int: INT,
    long: LONG,
    float: FLOAT,
    unicode: UNICODE,
    InputType: opaque_taint,
    OutputType: opaque_taint,
    #types.MethodType: None, #None means that there is no taint equivalent for that
    #types.FunctionType: None,
    #types.InstanceType: None,
    }


def mapt(o, f_taint_aware, f_built_in = lambda o : o):
    oType=o.__class__
    if oType in ignorableTypes:
        return o
    elif oType in tclasses.values():
        return f_taint_aware(o)
    elif oType in tclasses.keys():
         return f_built_in(o)
    elif isinstance(o, list): #list
         return [mapt(x, f_taint_aware, f_built_in) for x in o]
    elif isinstance(o, tuple): #tuple
         return tuple(mapt(x, f_taint_aware, f_built_in) for x in o)
    elif isinstance(o, set): #set
         return  set(mapt(x, f_taint_aware, f_built_in) for x in o)
    elif issubclass(oType,db.Model): #it's a tModel
        properties=o.properties()
        del properties['taints']
        return  dict((k, mapt(getattr(o,k), f_taint_aware, f_built_in)) for k in properties.keys())
    elif isinstance(o, dict): #dict
          klass = type(o) # It's quite common for frameworks to extend dict
                          # with useful new methdos - i.e. web.py
          return klass((k, mapt(v, f_taint_aware, f_built_in)) for k, v in o.iteritems()) #CHECKME esta bien que manche solo el valor y no la key?
    elif isinstance(o, types.InstanceType) or isinstance(o,object):
        for name in filter(lambda n: n[0]!='_',dir(o)):
            value=getattr(o,name)
            if not callable(value):
                newvalue=mapt(value, f_taint_aware, f_built_in)
                setattr(o,name,newvalue) #FIXME this overwrite all the attributes in every access. Maybe there is other way to do this.
        return o
    else:
        # with isinstance(o,object) in the previous case, we would not enter in this branch anymore... Should we remove it? FIXME
        raise Exception("I cannot apply mapt to the given object!\n\tObject:%s\n\tType: %s\n\t __class__: %s)" % (o,type(o),o.__class__))

def remove_tags(o, v):
    mapt(o, lambda t: (t.taints.downgrader(v), t)[1])
    return None

def collect_tags(o):
    '''Collect tags from a source s into a target t.'''
    t = taintConfig.Policy()
    if hasattr(o, 'taints'):
        mapt(o, lambda o: (t.join(o.taints), o)[1])
    return t

def update_tags(o, t):
    # Only should be run when you are sure that o is taint aware
    mapt(o, lambda o: (o.taints.join(t), o)[1])
    return None

def untclass(o):
    '''returns the plain flavor of a tainted thing'''
    if hasattr(o,'originalType'):
        return o.originalType(o)
    else:
        raise KeyError

def untaint_aware(r):
    # untaint r and returns the original object (sort of)
    r = mapt(r, untclass, lambda o: o)
    return r

from taintmode import untrusted

def tclass(o):
    '''Tainted instance factory.'''
    klass = type(o)
    if klass in tclasses.keys():
        #if klass == types.InstanceType: return taint_hook(o)
        if callable(o): return untrusted(o)
        else:
            return tclasses[klass](o)
    else:
        raise KeyError

def taint_aware_with_top(r):
    return taint_aware(r, taintConfig.Policy(taintConfig.Policy.top))

def taint_aware(r, tag=None):
    if tag is None: tag=taintConfig.Policy(taintConfig.Policy.bottom)
    # taint r with ts tags
    if not isinstance(tag,taintConfig.Policy): raise Exception("%s is not a TAG" % type(tag))
    r = mapt(r, lambda o: o, tclass)
    update_tags(r, tag)
    return r
